mod bahnhof;
mod vvs;

use std::{collections::BTreeMap, sync::Arc, time::Duration};

use chrono::{Datelike, Local, TimeDelta, Timelike};
use rocket::{get, http::Status, response::content::RawHtml, routes, FromForm, State};
use tokio::{
    sync::Mutex,
    time::{sleep, Instant},
};

#[rocket::launch]
async fn rocket() -> _ {
    rocket::build()
        .manage(Arc::new(()))
        .manage(Mutex::new(
            Option::<(
                Instant,
                Result<bahnhof::DeparturesMerklingen, String>,
                String,
            )>::None,
        ))
        .mount(
            "/",
            routes![
                index,
                api_doc,
                vvs_find_stop_simple,
                vvs_mins_before_simple,
                vvs_mins_before_check_simple,
                vvs_short_in_hrs_mins_before_simple,
                vvs_short_in_hrs_mins_before_check_simple,
            ],
        )
}

#[get("/")]
async fn index(
    departures: &State<
        Mutex<
            Option<(
                Instant,
                Result<bahnhof::DeparturesMerklingen, String>,
                String,
            )>,
        >,
    >,
) -> RawHtml<String> {
    let mut departures = departures.lock().await;
    if departures.as_ref().is_none_or(|(last_updated, result, _)| {
        last_updated.elapsed() > Duration::from_secs(30 * if result.is_ok() { 5 } else { 1 })
    }) {
        *departures = Some((
            Instant::now(),
            bahnhof::departures_merklingen().await,
            tokio::fs::read_to_string("index.html")
                .await
                .unwrap_or_else(|_| include_str!("index.html").to_owned()),
        ));
    }
    let departures_str = match &*departures {
        Some((_, Ok(departures), _)) => {
            let mut departures_str = String::new();
            departures_str.push_str(r#"<table style="width:100%;">"#);
            let mut table_rows = BTreeMap::<_, (String, String, usize, usize)>::new();
            for departure in &departures.entries {
                let mut messages = vec![];
                for departure in departure.iter() {
                    if let Some(message) = &departure.messages {
                        for (_message_type, message) in message {
                            for message in message {
                                if let Some(text) = message.text.clone() {
                                    if message.important.is_some_and(|v| v) {
                                        if !messages.contains(&text) {
                                            messages.push(text);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                if let Some(departure) = departure.first() {
                    if let Some(time) = departure.time_delayed.or(departure.time_schedule) {
                        let row = table_rows
                            .entry((time.date_naive(), time.hour(), time.minute() >= 30))
                            .or_default();
                        let line_name = departure
                            .line_name
                            .as_ref()
                            .map(|v| v.as_str())
                            .unwrap_or("[RE?]");
                        let is_ulm = {
                            if let Some(destination) = departure.destination.as_ref() {
                                if let Some(name) = &destination.name {
                                    let name = name.to_lowercase();
                                    if name.starts_with("ulm") {
                                        Some(true)
                                    } else if name.starts_with("wendlingen")
                                        || name.starts_with("plochingen")
                                        || name.starts_with("stuttgart")
                                    {
                                        Some(false)
                                    } else {
                                        None
                                    }
                                } else {
                                    None
                                }
                            } else {
                                None
                            }
                        }
                        .or_else(|| {
                            match departure.platform_schedule.as_ref().map(|v| v.as_str()) {
                                Some("1") => Some(true),
                                Some("4") => Some(false),
                                _ => None,
                            }
                        });
                        let mut dep_str = "<div>".to_owned();
                        dep_str.push_str(html_escape::encode_safe(line_name).as_ref());
                        let platform = departure
                            .platform
                            .as_ref()
                            .map(|v| v.as_str())
                            .or(departure.platform_schedule.as_ref().map(|v| v.as_str()));
                        dep_str.push_str(&format!(
                            " <small{}>{:0>2}:{:0>2}</small>{}{}",
                            if let Some(schedule) = &departure.time_schedule {
                                let delay = ({
                                    if time > *schedule {
                                        time - *schedule
                                    } else {
                                        TimeDelta::zero()
                                    }
                                }
                                .num_minutes() as f64
                                    / 15.0)
                                    .max(0.0)
                                    .min(1.0);
                                format!(
                                    r#" style="color: hsl({}, 60%, 60%)""#,
                                    (120u8.saturating_sub((120.0 * delay) as u8))
                                )
                            } else {
                                String::new()
                            },
                            time.hour(),
                            time.minute(),
                            if let Some(platform) = platform {
                                if platform.trim().is_empty() {
                                    ""
                                } else if departure
                                    .platform_schedule
                                    .as_ref()
                                    .is_some_and(|scheduled| platform != scheduled)
                                {
                                    r#"<small style="color:darkorange;"> auf Gleis "#
                                } else {
                                    r#"<small style="color:gray"> auf Gleis "#
                                }
                            } else {
                                ""
                            },
                            if let Some(platform) = platform {
                                if platform.trim().is_empty() {
                                    String::new()
                                } else {
                                    format!("{platform}</small>")
                                }
                            } else {
                                String::new()
                            },
                        ));
                        if let Some(dest) = &departure.destination {
                            if let Some(name) = &dest.name {
                                dep_str.push_str(r#"<br><span style="color:gray;"> → </span>"#);
                                dep_str.push_str(html_escape::encode_safe(name).as_ref());
                            }
                        }
                        if departure.canceled.is_some_and(|canceled| canceled) {
                            dep_str.push_str(
                                r#"<div style="color:darkorange;"> ▸ fällt aus / cancelled</div>"#,
                            );
                        } else if let Some(stop_place) =
                            departure.stop_place.as_ref().filter(|stop_place| {
                                stop_place.canceled.is_some_and(|canceled| canceled)
                            })
                        {
                            dep_str.push_str(r#"<div style="color:darkorange;"> ▸ stop/halt "#);
                            html_escape::encode_safe_to_string(
                                stop_place
                                    .name
                                    .as_ref()
                                    .map(|v| v.as_str())
                                    .unwrap_or("[stopPlace.name]"),
                                &mut dep_str,
                            );
                            dep_str.push_str(": fällt aus / cancelled</div>");
                        }
                        for message in &messages {
                            dep_str.push_str(r#"<div style="color:darkorange;"> ▹ "#);
                            html_escape::encode_safe_to_string(message, &mut dep_str);
                            dep_str.push_str("</div>");
                        }
                        dep_str.push_str("</div>");
                        match is_ulm {
                            Some(true) => {
                                row.0.push_str(&dep_str);
                                row.2 += 1;
                            }
                            Some(false) => {
                                row.1.push_str(&dep_str);
                                row.3 += 1;
                            }
                            None => {
                                if row.2 <= row.3 {
                                    row.0.push_str(&dep_str);
                                    row.2 += 1;
                                } else {
                                    row.1.push_str(&dep_str);
                                    row.3 += 1;
                                }
                            }
                        }
                    }
                }
            }
            for ((_date, _hour, _late_half), (a, b, _, _)) in table_rows {
                departures_str.push_str("<tr><td>");
                departures_str.push_str(&a);
                departures_str.push_str("</td><td>");
                departures_str.push_str(&b);
                departures_str.push_str("</td></tr>");
            }
            departures_str.push_str("</table>");
            departures_str
        }
        Some((_, Err(e), _)) => format!("<small>{}</small>", html_escape::encode_safe(e)),
        None => "<small>failed</small>".to_owned(),
    };
    RawHtml(
        departures
            .as_ref()
            .map(|(_, _, index)| index.as_str())
            .unwrap_or(include_str!("index.html"))
            .replace("%%DEPARTURES%%", &departures_str),
    )
}

#[get("/api")]
async fn api_doc() -> RawHtml<String> {
    let tomorrow = Local::now() + Duration::from_secs(60 * 60 * 24);
    RawHtml(
        include_str!("api.html")
            .replace("%YEAR%", &tomorrow.year().to_string())
            .replace("%MONTH%", &tomorrow.month().to_string())
            .replace("%DAY%", &tomorrow.day().to_string()),
    )
}

async fn rate_limit(rate_limit: &State<Arc<()>>) -> Arc<()> {
    let rate_limit = Arc::clone(rate_limit);
    let secs = (Arc::strong_count(&rate_limit).saturating_sub(3) as u64).saturating_mul(2);
    if secs > 0 {
        sleep(Duration::from_secs(secs)).await;
    }
    rate_limit
}

#[get("/vvs/find_stop/simple/<stop>")]
async fn vvs_find_stop_simple(stop: &str, counter: &State<Arc<()>>) -> String {
    let _counter = rate_limit(counter).await;
    match vvs::find_stop_best(stop).await {
        Ok(Some((name, _))) => {
            format!("ok: {name}")
        }
        Ok(None) => format!("err: not found"),
        Err(e) => format!("err: {e}"),
    }
}

#[derive(FromForm)]
struct VvsShortInHrsMinsBeforeSimple {
    hour: Option<u8>,
    minute: Option<u8>,
    interchanges: Option<u16>,
    allowtaxi: Option<bool>,
}
#[get("/vvs/short/in_hrs/<hrs>/mins_before/<from>/<to>?<form..>")]
async fn vvs_short_in_hrs_mins_before_simple(
    hrs: f64,
    from: &str,
    to: &str,
    form: VvsShortInHrsMinsBeforeSimple,
    counter: &State<Arc<()>>,
) -> (Status, String) {
    let time = Local::now() + Duration::from_secs_f64(hrs * 60.0 * 60.0);
    let form = VvsMinsBeforeSimpleForm {
        year: time.year() as _,
        month: time.month() as _,
        day: time.day() as _,
        hour: form.hour.unwrap_or(time.hour() as _),
        minute: form.minute.unwrap_or(time.minute() as _),
        interchanges: form.interchanges,
        allowtaxi: form.allowtaxi,
    };
    vvs_mins_before_simple_impl(from, to, form, counter).await
}
#[derive(FromForm)]
struct VvsShortInHrsMinsBeforeCheckSimple {
    hour: Option<u8>,
    minute: Option<u8>,
    interchanges: Option<u16>,
    minutes: u32,
    allowtaxi: Option<bool>,
}
#[get("/vvs/short/in_hrs/<hrs>/mins_before_check/<from>/<to>?<form..>")]
async fn vvs_short_in_hrs_mins_before_check_simple(
    hrs: f64,
    from: &str,
    to: &str,
    form: VvsShortInHrsMinsBeforeCheckSimple,
    counter: &State<Arc<()>>,
) -> (Status, String) {
    let time = Local::now() + Duration::from_secs_f64(hrs * 60.0 * 60.0);
    let form = VvsMinsBeforeCheckSimpleForm {
        year: time.year() as _,
        month: time.month() as _,
        day: time.day() as _,
        hour: form.hour.unwrap_or(time.hour() as _),
        minute: form.minute.unwrap_or(time.minute() as _),
        interchanges: form.interchanges,
        minutes: form.minutes,
        allowtaxi: form.allowtaxi,
    };
    vvs_mins_before_check_simple_impl(from, to, form, counter).await
}
#[derive(FromForm)]
struct VvsMinsBeforeSimpleForm {
    year: u16,
    month: u8,
    day: u8,
    hour: u8,
    minute: u8,
    interchanges: Option<u16>,
    allowtaxi: Option<bool>,
}
#[get("/vvs/mins_before/simple/<from>/<to>?<form..>")]
async fn vvs_mins_before_simple(
    from: &str,
    to: &str,
    form: VvsMinsBeforeSimpleForm,
    counter: &State<Arc<()>>,
) -> (Status, String) {
    vvs_mins_before_simple_impl(from, to, form, counter).await
}
async fn vvs_mins_before_simple_impl(
    from: &str,
    to: &str,
    form: VvsMinsBeforeSimpleForm,
    counter: &State<Arc<()>>,
) -> (Status, String) {
    let _counter = rate_limit(counter).await;
    if Arc::strong_count(&counter) > 2 {}
    let (_, from) = match vvs::find_stop_best(from).await {
        Ok(Some(v)) => v,
        Ok(None) => {
            return (
                Status::NotFound,
                format!("err: Couldn't find origin stop {from}"),
            )
        }
        Err(e) => return (Status::InternalServerError, format!("err: {e}")),
    };
    let (_, to) = match vvs::find_stop_best(to).await {
        Ok(Some(v)) => v,
        Ok(None) => {
            return (
                Status::NotFound,
                format!("err: Couldn't find destination stop {to}"),
            )
        }
        Err(e) => return (Status::InternalServerError, format!("err: {e}")),
    };
    match vvs::find_trip_arrival_before(
        &from,
        &to,
        form.year,
        form.month,
        form.day,
        form.hour,
        form.minute,
        form.interchanges,
        form.allowtaxi.is_some_and(|allow| allow),
    )
    .await
    {
        Ok(Some(v)) => (Status::Ok, format!("minutes: {v}")),
        Ok(None) => (
            Status::Gone,
            format!("gone: maybe try with different parameters or check vvs.de"),
        ),
        Err(e) => return (Status::InternalServerError, format!("err: {e}")),
    }
}
#[derive(FromForm)]
struct VvsMinsBeforeCheckSimpleForm {
    year: u16,
    month: u8,
    day: u8,
    hour: u8,
    minute: u8,
    interchanges: Option<u16>,
    minutes: u32,
    allowtaxi: Option<bool>,
}
#[get("/vvs/mins_before_check/simple/<from>/<to>?<form..>")]
async fn vvs_mins_before_check_simple(
    from: &str,
    to: &str,
    form: VvsMinsBeforeCheckSimpleForm,
    counter: &State<Arc<()>>,
) -> (Status, String) {
    vvs_mins_before_check_simple_impl(from, to, form, counter).await
}
async fn vvs_mins_before_check_simple_impl(
    from: &str,
    to: &str,
    form: VvsMinsBeforeCheckSimpleForm,
    counter: &State<Arc<()>>,
) -> (Status, String) {
    let _counter = rate_limit(counter).await;
    let (_, from) = match vvs::find_stop_best(from).await {
        Ok(Some(v)) => v,
        Ok(None) => {
            return (
                Status::NotFound,
                format!("err: Couldn't find origin stop {from}"),
            )
        }
        Err(e) => return (Status::InternalServerError, format!("err: {e}")),
    };
    let (_, to) = match vvs::find_stop_best(to).await {
        Ok(Some(v)) => v,
        Ok(None) => {
            return (
                Status::NotFound,
                format!("err: Couldn't find destination stop {to}"),
            )
        }
        Err(e) => return (Status::InternalServerError, format!("err: {e}")),
    };
    match vvs::find_trip_arrival_before(
        &from,
        &to,
        form.year,
        form.month,
        form.day,
        form.hour,
        form.minute,
        form.interchanges,
        form.allowtaxi.is_some_and(|allow| allow),
    )
    .await
    {
        Ok(Some(v)) => {
            if v <= form.minutes {
                (Status::Ok, String::new())
            } else {
                (
                    Status::Gone,
                    format!("gone: {v} > {} minutes", form.minutes),
                )
            }
        }
        Ok(None) => (
            Status::Gone,
            format!("gone: maybe try with different parameters or check vvs.de"),
        ),
        Err(e) => return (Status::InternalServerError, format!("err: {e}")),
    }
}
