use chrono::{DateTime, Local, TimeZone};
use serde::Deserialize;

#[derive(Debug)]
pub struct VvsStopIdentifier {
    id: String,
    gid: String,
}

pub async fn find_stop_best(query: &str) -> Result<Option<(String, VvsStopIdentifier)>, String> {
    Ok(find_stop(query)
        .await?
        .into_iter()
        .filter_map(|(name, _types, identifier)| Some((name, identifier?)))
        .next())
}

pub async fn find_stop(
    query: &str,
) -> Result<
    Vec<(
        String,
        (Option<String>, Option<String>),
        Option<VvsStopIdentifier>,
    )>,
    String,
> {
    let url = format!(
        r#"https://www3.vvs.de/mngvvs/XML_STOPFINDER_REQUEST
                ?SpEncId=0
                &coordOutputFormat=EPSG:4326
                &serverInfo=1
                &suggestApp=vvs
                &type_sf=any
                &version=10.2.10.139
                &jsonp=func
                &suggest_macro=vvs
                &name_sf={}
        "#,
        urlencoding::encode(query)
    )
    .replace(['\r', '\n', ' ', '\t'], "");
    match reqwest::get(&url).await {
        Ok(response) => match response.text().await {
            Ok(response) => {
                let response = response
                    .trim_start_matches("func(")
                    .trim_end_matches([')', ';']);
                #[derive(Deserialize)]
                struct Response0a {
                    #[serde(rename = "stopFinder")]
                    stop_finder: Response1a,
                }
                #[derive(Deserialize)]
                struct Response0b {
                    #[serde(rename = "stopFinder")]
                    stop_finder: Response1b,
                }
                #[derive(Deserialize)]
                struct Response1a {
                    #[serde(rename = "points")]
                    points: Vec<Response2>,
                }
                #[derive(Deserialize)]
                struct Response1b {
                    #[serde(rename = "points")]
                    points: Response2Container,
                }
                #[derive(Deserialize)]
                struct Response2Container {
                    #[serde(rename = "point")]
                    point: Response2,
                }
                #[derive(Deserialize)]
                struct Response2 {
                    #[serde(rename = "type")]
                    type1: Option<String>,
                    #[serde(rename = "anyType")]
                    type2: Option<String>,
                    #[serde(rename = "name")]
                    name: String,
                    #[serde(rename = "ref")]
                    identifier: Response3,
                }
                #[derive(Deserialize)]
                struct Response3 {
                    #[serde(rename = "id")]
                    id: String,
                    #[serde(rename = "gid")]
                    gid: String,
                }
                match serde_json::from_str::<Response0a>(&response)
                    .or_else(|ea| serde_json::from_str::<Response0b>(&response)
                        .map(|response_b| Response0a {
                            stop_finder: Response1a {
                                points:vec![response_b.stop_finder.points.point]
                            }
                        })
                        .map_err(|eb| (ea, eb))
                    )
                {
                    Ok(response) => Ok(response
                        .stop_finder
                        .points
                        .into_iter()
                        .map(|point| {
                            (
                                point.name,
                                (point.type1, point.type2),
                                if !(point.identifier.id.is_empty() || point.identifier.gid.is_empty() || point.identifier.id == "-1" || point.identifier.gid == "-1") {
                                    Some(VvsStopIdentifier {
                                        id: point.identifier.id,
                                        gid: point.identifier.gid,
                                    })
                                } else { None }
                            )
                        })
                        .collect()),
                    Err((ea, eb)) => {
                        Err(format!("Couldn't parse HTTP response from URL {url:?}: {ea} & {eb}\nResponse was (raw):\n{response}"))?
                    }
                }
            }
            Err(e) => Err(format!("Couldn't get HTTP response from URL {url:?}: {e}"))?,
        },
        Err(e) => Err(format!("Couldn't make GET request to URL {url:?}: {e}"))?,
    }
}

/// requests trips arriving before or at the specified time.
/// of the trips the api returned, ignores all that have more than `max_interchanges` interchanges, if `max_interchanges.is_some()`.
/// of the remaining trips, calculates how many minutes before the requested arrival time the departure time is, and returns the minimum of these durations in minutes.
/// For example, if you request to arrive at 8:50, a trip arrives at 8:45 and departs at 8:40, then this function would return `Some(10)`.
/// Returns `None` if no valid trips were found.
pub async fn find_trip_arrival_before(
    from: &VvsStopIdentifier,
    to: &VvsStopIdentifier,
    year: u16,
    month: u8,
    day: u8,
    hour: u8,
    minute: u8,
    max_interchanges: Option<u16>,
    allow_taxi: bool,
) -> Result<Option<u32>, String> {
    if year > 9999 {
        Err(format!(
            "Year {year} is out of range (cannot be represented with length 4)"
        ))?;
    }
    if month > 99 {
        Err(format!(
            "Month {month} is out of range (cannot be represented with length 2)"
        ))?;
    }
    if day > 99 {
        Err(format!(
            "Day {day} is out of range (cannot be represented with length 2)"
        ))?;
    }
    if hour > 99 {
        Err(format!(
            "Hour {hour} is out of range (cannot be represented with length 2)"
        ))?;
    }
    if minute > 99 {
        Err(format!(
            "Minute {minute} is out of range (cannot be represented with length 2)"
        ))?;
    }
    let url = format!(
        r#"
            https://www3.vvs.de/mngvvs/XML_TRIP_REQUEST2
            ?SpEncId=0
            &changeSpeed=normal
            &computationType=sequence
            &coordOutputFormat=EPSG:4326
            &cycleSpeed=14
            &deleteAssignedStops=0
            &deleteITPTWalk=0
            &descWithElev=1
            &illumTransfer=on
            &itOptionsActive=1
            &itdDate={year:0>4}{month:0>2}{day:0>2}
            &itdTime={hour:0>2}{minute:0>2}
            &itdTripDateTimeDepArr=arr
            &language=de
            &locationServerActive=1
            &macroWebTrip=true
            &name_destination={}
            &name_origin={}
            &noElevationProfile=1
            &noElevationSummary=1
            &outputFormat=rapidJSON
            &outputOptionsActive=1
            &ptOptionsActive=1
            &routeType=leasttime
            &searchLimitMinutes=360
            &securityOptionsActive=1
            &serverInfo=1
            &trITArrMOT=100
            &trITArrMOTvalue=15
            &trITDepMOT=100
            &trITDepMOTvalue=15
            &tryToFindLocalityStops=1
            &type_destination=any
            &type_origin=any
            &useElevationData=1
            &useLocalityMainStop=0
            &useRealtime=1
            &useUT=1
            &version=10.2.10.139
            &w_objPrefAl=12
            &w_regPrefAm=1
        "#,
        urlencoding::encode(&to.gid),
        urlencoding::encode(&from.gid)
    )
    .replace(['\r', '\n', ' ', '\t'], "");
    match reqwest::get(&url).await {
        Ok(response) => match response.text().await {
            Ok(response) => {
                #[derive(Deserialize)]
                struct Response0 {
                    #[serde(rename = "journeys")]
                    journeys: Vec<Response1>,
                }
                #[derive(Deserialize)]
                struct Response1 {
                    #[serde(rename = "interchanges")]
                    interchanges: i32,
                    #[serde(rename = "legs")]
                    legs: Vec<Response2>,
                }
                #[derive(Deserialize)]
                struct Response2 {
                    /// in seconds
                    #[serde(rename = "duration")]
                    duration: i32,
                    #[serde(rename = "origin")]
                    origin: Response3,
                    #[serde(rename = "destination")]
                    destination: Response4,
                    #[serde(rename = "transportation")]
                    transportation: Response5,
                }
                #[derive(Deserialize)]
                struct Response3 {
                    #[serde(rename = "name")]
                    name: String,
                    #[serde(rename = "departureTimeBaseTimetable")]
                    pub departure_time_base_timetable: Option<DateTime<Local>>,
                    #[serde(rename = "departureTimePlanned")]
                    pub departure_time_planned: Option<DateTime<Local>>,
                    #[serde(rename = "departureTimeEstimated")]
                    pub departure_time_estimated: Option<DateTime<Local>>,
                }
                #[derive(Deserialize)]
                struct Response4 {
                    #[serde(rename = "name")]
                    name: String,
                    #[serde(rename = "arrivalTimeBaseTimetable")]
                    pub arrival_time_base_timetable: Option<DateTime<Local>>,
                    #[serde(rename = "arrivalTimePlanned")]
                    pub arrival_time_planned: Option<DateTime<Local>>,
                    #[serde(rename = "arrivalTimeEstimated")]
                    pub arrival_time_estimated: Option<DateTime<Local>>,
                }
                #[derive(Deserialize)]
                struct Response5 {
                    #[serde(rename = "name")]
                    name: String,
                    #[serde(rename = "number")]
                    number: String,
                }
                match serde_json::from_str::<Response0>(&response) {
                    Ok(response) => Ok(response
                        .journeys
                        .iter()
                        .rev()
                        .filter_map(|journey| {
                            if let (Some(first_leg), Some(last_leg)) =
                                (journey.legs.first(), journey.legs.last())
                            {
                                if !allow_taxi && journey.legs.iter().any(|leg| leg.transportation.name.to_lowercase().contains("taxi") || leg.transportation.number.to_lowercase().contains("taxi")) {
                                    return None;
                                }
                                if max_interchanges
                                    .is_none_or(|max| journey.interchanges <= max as i32)
                                {
                                    if let Some(arrival_time) = last_leg
                                        .destination
                                        .arrival_time_estimated
                                        .as_ref()
                                        .or(last_leg.destination.arrival_time_planned.as_ref())
                                        .or(last_leg
                                            .destination
                                            .arrival_time_base_timetable
                                            .as_ref())
                                    {
                                        if *arrival_time
                                            <= Local
                                                .with_ymd_and_hms(
                                                    year as _,
                                                    month as _,
                                                    day as _,
                                                    hour as _,
                                                    minute as _,
                                                    59,
                                                )
                                                .latest()?
                                        {
                                            if let Some(departure_time) = first_leg
                                                .origin
                                                .departure_time_estimated
                                                .as_ref()
                                                .or(last_leg.origin.departure_time_planned.as_ref())
                                                .or(last_leg
                                                    .origin
                                                    .departure_time_base_timetable
                                                    .as_ref())
                                            {
                                                let arrival_wanted = Local
                                                    .with_ymd_and_hms(
                                                        year as _,
                                                        month as _,
                                                        day as _,
                                                        hour as _,
                                                        minute as _,
                                                        0,
                                                    )
                                                    .latest()?;
                                                if *departure_time <= arrival_wanted {
                                                    Some(
                                                        (arrival_wanted - departure_time)
                                                            .num_minutes()
                                                            as _,
                                                    )
                                                } else {
                                                    None
                                                }
                                            } else {
                                                None
                                            }
                                        } else {
                                            None
                                        }
                                    } else {
                                        None
                                    }
                                } else {
                                    None
                                }
                            } else {
                                None
                            }
                        })
                        .min()),
                    Err(e) => Err(format!("Couldn't parse HTTP response from URL {url:?}: {e}\nResponse was (raw):\n{response}"))?,
                }
            }
            Err(e) => Err(format!("Couldn't get HTTP response from URL {url:?}: {e}"))?,
        },
        Err(e) => Err(format!("Couldn't make GET request to URL {url:?}: {e}"))?,
    }
}
